;; ----------------------------------------------------------------------------
;; .emacs -- Snippets from from many emacsen files.
;;
;; Documentation modes:
;; - Markdown
;; - Org
;;
;; Programming languages modes with IDE like support:
;; - C/C++
;; - elisp
;; - Go
;; - OCaml
;;
;; Simple programming language modes:
;; - Nim
;; - Zig
;;
;; Other features:
;; - Auto complete (auto-complete, company)
;; - Compiler diagnostics (flycheck)
;; - Extraneous whitespace indications
;; - Find file (ido)
;; - GDB support (gud)
;; - Git support (magit)
;; - Project like IDE support (treemacs, projectile, lsp-mode)
;; - Templated auto complete assisted snippets (yasnippet)
;;
;; Requires:
;;  Emacs >= 25.2
;; ----------------------------------------------------------------------------

;; See https://www.gnu.org/software/emacs/manual/html_node/elisp/Key-Binding-Conventions.html
;; for user key binding conventions.

;; ----------------------------------------------------------------------------
;; User key bindings:
;; ----------------------------------------------------------------------------
;;
;; C-c j   goto-line (could remove as Emacs now ships with "M-g g" for goto-line)
;; C-c s   sort-lines
;; C-c a   align-current
;; C-c i   ispell-region
;;
;; C-c g   find-grep
;; C-c c   compile
;; C-c t   treemacs
;;
;; C-c k   auto complete          (in progamming modes)
;; C-c o   ff-find-other-file     (in c-mode-common)
;; C-c O   projectile-find-file   (find file in project or similar function)
;; C-c l   lsp-mode command prefix
;; C-c d   lsp-ui-doc-glance      (... or display doc string at cursor in general)
;; C-c m   lsp-ui-menu            (display side menu)
;;
;; C-c k   isearch mark and exit
;; C-c o   isearch-occur
;; C-c SPC isearch-toggle-lax-whitespace
;;
;; C-c n   org-capture (quick note)
;; C-c p   org-agenda (planning)
;;
;; f5     GDB: stop
;; f6     GDB: continue
;; f7     GDB: step
;; S-f7   GDB: next -- Does not work in PuTTY
;; f8     GDB: finish
;;
;; f9     next-error (or grep result etc)
;; S-f9   prev-error (or grep result etc) -- Does not work in PuTTY
;; f10    prev-error (or grep result etc) -- Steal "global menu" key for prev-error alias

;; ----------------------------------------------------------------------------
;; Key overrides (modifies standard Emacs behavior or use non-user prefix).
;; ----------------------------------------------------------------------------
;;
;; M-n    scroll-up-command   (replaces: -)
;; M-p    scroll-down-command (replaces: -)
;; M-f    forward-symbol      (replaces: forward-word)
;; M-b    backward-symbol     (replaces: backward-word)
;; M-C-f  forward-paragraph   (replaces: forward-sexp)
;; M-C-b  backward-paragraph  (replaces: backward-sexp)
;; C-x k  fast-kill-buffer    (replaces: kill-buffer)

;; ----------------------------------------------------------------------------
;; Convenience functions defined by this file (not loaded modes):
;; ----------------------------------------------------------------------------
;;
;; reload-file       revert-buffer without confirmation
;; fast-kill-buffer  kill-buffer without confirmation

;; ----------------------------------------------------------------------------
;; Paths to external tools, change according to taste.
;; ----------------------------------------------------------------------------

(defvar my/grep-find-command "~/.emacs.d/bin/csearch-color "
  "Grep find command.")

(defvar my/ccls-exe-path (executable-find "ccls")
  "Path to ccls executable (C++ LSP indexer)")

;; ----------------------------------------------------------------------------
;; Straight package manager
;;   https://github.com/raxod502/straight.el
;;
;;   Packages loaded by this file is managed by the straight package manager.
;;
;; Requires:
;;   emacs >= 24.5
;;   git
;; ----------------------------------------------------------------------------

;; This bootstraps the package manager if it does not already exist in ~/.emacs.d/...
(defvar bootstrap-version)
(let ((bootstrap-file
       (expand-file-name "straight/repos/straight.el/bootstrap.el" user-emacs-directory))
      (bootstrap-version 5))
  (unless (file-exists-p bootstrap-file)
    (with-current-buffer
        (url-retrieve-synchronously
         "https://raw.githubusercontent.com/raxod502/straight.el/develop/install.el"
         'silent 'inhibit-cookies)
      (goto-char (point-max))
      (eval-print-last-sexp)))
  (load bootstrap-file nil 'nomessage))

;; About use-package coexistence with straight:
;;
;;   The ":ensure t" clause does not work as it is hard-coded to install
;;   packages using package.el. Use ":ensure nil" instead and an explicit
;;   straight-use-package before the use-package. See
;;   https://github.com/raxod502/straight.el/issues/128 for details.
(straight-use-package 'use-package)

;; Various support functions
(straight-use-package 'auto-complete)
(straight-use-package 'company)
(straight-use-package 'company-lsp)
(straight-use-package 'yasnippet)
(straight-use-package 'flycheck)
(straight-use-package 'treemacs)
(straight-use-package 'projectile)
(straight-use-package 'lsp-mode)
(straight-use-package 'lsp-ui)
(straight-use-package 'lsp-treemacs)
(straight-use-package 'which-key)

;; Themes
(straight-use-package 'solarized-theme)

;; Documentation modes
(straight-use-package 'markdown-mode)
(straight-use-package 'org)

;; Language modes
(straight-use-package 'go-mode)
(straight-use-package 'nim-mode)
(straight-use-package 'zig-mode)
(straight-use-package 'ccls) ;; C/C++ LSP client

;; GIT plugin
(straight-use-package 'magit)

;; For LISP code
(straight-use-package 'paredit)

;; ----------------------------------------------------------------------------
;; Misc minor tweaks
;; ----------------------------------------------------------------------------

;; Disable pager for "git grep"
(setenv "PAGER" "cat")

;; Local emacs config dir.
(add-to-list 'load-path "~/.emacs.d/local" t)
(add-to-list 'custom-theme-load-path "~/.emacs.d/local/themes/emacs-color-theme-solarized")

;; Emacs <= 26.3 does not recognize st-256color terminal.
(add-to-list 'term-file-aliases
    '("st-256color" . "xterm-256color"))

;; Disable startup message
(setq inhibit-startup-message t)

;; Make all "yes or no" prompts show "y or n" instead
(fset 'yes-or-no-p 'y-or-n-p)

;; Don't create .saves-xxx files
(setq auto-save-list-file-prefix nil)

;; Open unidentified files in fundamental mode
(setq default-major-mode 'fundamental-mode)

;; Use UTF8 character encoding.
(prefer-coding-system 'utf-8)

;; No new lines when moving past end of buffer
(setq next-line-add-newlines nil)

;; Display marked region
(setq transient-mark-mode t)

;; Allow upcase, downcase commands (we are not morons).
(put 'upcase-region 'disabled nil)
(put 'downcase-region 'disabled nil)

;; Set fill column used by fill-region.
;; Override value with M-x f
(setq-default fill-column 80)

;; Turn off stupid double space rules!
(setq sentence-end-double-space nil)

;; Follow symlinks to version controlled controlled files.
(setq vc-follow-symlinks t)

;; Somehow this is needed for ediff to be able to diff files with local modifications in git.
(setq vc-handled-backends nil)

;; Display both line number and column number
(line-number-mode 1)
(column-number-mode 1)

;; Show matching parenthesis
(show-paren-mode 1)

;; Show the function we are in when editing source code.
(which-function-mode 1)

;; Compilation output
(setq compilation-scroll-output t)

;; Turn off menu and tool bar by default
(menu-bar-mode -1)
(tool-bar-mode -1)
(scroll-bar-mode -1)

;; multi-line help text in the echo area
(setq tooltip-use-echo-area t)

;; Turn on auto fill mode when editing text files.
;;(add-hook 'text-mode-hook 'turn-on-auto-fill)

(setq split-height-threshold nil)
(setq split-width-threshold 160)

;; Keep some context above cursor when scrolling pages.
(setq scroll-margin 5)

;; Not a big fan of pulse animations, find it slightly distracting.
;; The block cursor is enough to find my bearing most of the time. Set a short
;; interval as a compromise as it may be hard to see an unfocused block cursor
;; when using an imenu side bar for navigation.
(setq pulse-iterations 3)

(defun reload-file ()
    "Revert buffer without confirmation."
    (interactive) (revert-buffer t t))

;; Prevent projectile from recursing down into Git sub modules as that may take
;; quite some time in large projects. Think of sub modules as separate projects.
(setq projectile-git-submodule-command "")

(defun my/isearch-mark-and-exit ()
  (interactive)
  (isearch-done)
  (push-mark isearch-other-end 'no-message 'activate))

(define-key isearch-mode-map (kbd "C-c k") 'my/isearch-mark-and-exit)
(define-key isearch-mode-map (kbd "C-c o") 'isearch-occur)
(define-key isearch-mode-map (kbd "C-c SPC") 'isearch-toggle-lax-whitespace)

;; ----------------------------------------------------------------------------
;; Interactive key binding reminder support
;; ----------------------------------------------------------------------------

(require 'which-key)
(which-key-mode)

;; ----------------------------------------------------------------------------
;; Key bindings
;; ----------------------------------------------------------------------------

;; Goto line number
(global-set-key (kbd "C-c j") 'goto-line)

;; Sort lines
(global-set-key (kbd "C-c s") 'sort-lines)

;; Align current block of code
(global-set-key (kbd "C-c a") 'align-current)

;; Spell check current selection
(global-set-key (kbd "C-c i") 'ispell-region)

;; Grep
(global-set-key (kbd "C-c g") 'find-grep)

;; Compile
(global-set-key (kbd "C-c c") 'compile)

;; Show or hide treemacs pane.
(global-set-key (kbd "C-c t") 'treemacs)

;; Find file in project.
(global-set-key (kbd "C-c O") 'projectile-find-file)

;; Capture quick note
(global-set-key (kbd "C-c n") 'org-capture)

;; Show current agenda
(global-set-key (kbd "C-c p") 'org-agenda)

;; Compile and find-grep results etc
(global-set-key (kbd "<f9>") 'next-error)
(global-set-key (kbd "<S-f9>") 'previous-error)
;; Shift + Fn keys does not work in putty.
;; Define additional key as a workaround.
(global-set-key (kbd "<f10>") 'previous-error)

;; Scrolling
(global-set-key (kbd "M-n") 'scroll-up-command)
(global-set-key (kbd "M-p") 'scroll-down-command)

(defun backward-symbol (&optional arg)
   "Move backward until encountering the beginning of a symbol.
With argument, do this that many times."
   (interactive "p")
    (forward-symbol (- (or arg 1))))

;; Move by symbols instead of words.
(global-set-key (kbd "M-f") 'forward-symbol)
(global-set-key (kbd "M-b") 'backward-symbol)

;; Forward, backward paragraph. Usually bound to ctrl + up/down but those key
;; combinations usually don't work in remote terminals.
(global-set-key (kbd "M-C-f") 'forward-paragraph)
(global-set-key (kbd "M-C-b") 'backward-paragraph)

;; Kill the default buffer without questions.
(defun fast-kill-buffer ()
   (interactive)
   (kill-buffer (buffer-name)))

;; Faster kill default buffer command.
(define-key global-map [C-delete] 'fast-kill-buffer)
(substitute-key-definition 'kill-buffer 'fast-kill-buffer (current-global-map))

;; ----------------------------------------------------------------------------
;; Font
;;   https://blog.golang.org/go-fonts
;;   https://fonts.google.com/specimen/Inconsolata
;; ----------------------------------------------------------------------------

(require 'cl-lib)
(defun font-candidate (&rest fonts)
  "Return existing font which first match."
  (cl-find-if (lambda (f) (find-font (font-spec :name f))) fonts))

(when (display-graphic-p)
  (set-face-attribute
   'default nil :font (font-candidate "Go Mono-9" "Inconsolata-10" "DejaVu Sans Mono-10")))

;; ----------------------------------------------------------------------------
;; Buffer management
;; ----------------------------------------------------------------------------

;; Use M-x list-buffers in the event you need the full buffer list in a separate frame
(global-set-key (kbd "C-x C-b") 'bs-show)

;; ----------------------------------------------------------------------------
;; Whitespace config
;; ----------------------------------------------------------------------------

(setq-default
 whitespace-line-column 120
 whitespace-action '(auto-cleanup)
 whitespace-style
 '(face empty trailing))

(add-hook 'prog-mode-hook 'whitespace-mode)

;; ----------------------------------------------------------------------------
;; Ido
;;   http://www.emacswiki.org/emacs/InteractivelyDoThings
;; ----------------------------------------------------------------------------

(require 'ido)

;; Prevent IDO from looking for non existing files in other directories.
;; I find it annoying as most of the time I'm about to create a new file.
(setq ido-auto-merge-work-directories-length -1)

(ido-mode t)

;; ----------------------------------------------------------------------------
;; GDB
;; ----------------------------------------------------------------------------

;; Activate by "M-x gdb" followed by "M-x gdb-many-windows".
;;
;; NOTE: The toolbar contain debugger navigation icons so you may want to enable
;;       that while debugging using "M-x tool-bar-mode".

;; Stop
(global-set-key (kbd "<f5>") (lambda ()
                               (interactive)
                               (gud-stop-subjob)
                               (comint-interrupt-subjob)))

;; Continue
(global-set-key (kbd "<f6>") 'gud-cont)

;; Step
(global-set-key (kbd "<f7>") 'gud-step)

;; Next
(global-set-key (kbd "<S-f7>") 'gud-next)

;; Finish
(global-set-key (kbd "<f8>") 'gud-finish)

;; ----------------------------------------------------------------------------
;; C mode (common)
;;
;; lsp-mode together with ccls is used for code navigation and auto complete
;; support. See ccls config further down for details.
;; ----------------------------------------------------------------------------

(defun my/c-mode-common-hook ()
  (c-set-style "Linux")
  (setq
   indent-tabs-mode nil
   c-basic-offset 4)
  (modify-syntax-entry ?_ "w" c-mode-syntax-table)
  (local-set-key  (kbd "C-c o") 'ff-find-other-file)
  ;; Disable auto-complete mode, use company mode.
  (auto-complete-mode -1))

(add-hook 'c-mode-common-hook 'my/c-mode-common-hook)

;; ----------------------------------------------------------------------------
;; C++ mode
;; ----------------------------------------------------------------------------

(defun my/c++-setup ()
  (c-set-offset 'innamespace [0]))

(add-hook 'c++-mode-hook 'my/c++-setup)

;; ----------------------------------------------------------------------------
;; csearch
;;   https://github.com/google/codesearch
;;
;; install support tools:
;;   cd ~/.emacs.d/src/github.com/google/codesearch/cmd
;;   for a in cgrep cindex csearch; do (cd $a && go install) done
;; ----------------------------------------------------------------------------

;; Simply override the grep-find command to call csearch instead.
(let ((cmd (bound-and-true-p my/grep-find-command)))
  (when cmd
    (setq grep-find-command cmd)))

;; ----------------------------------------------------------------------------
;; external package: solarized-theme (color theme)
;;   https://github.com/bbatsov/solarized-emacs
;; ----------------------------------------------------------------------------

;; Only activate the color theme in graphical Emacs.
;; This theme looks funky when running in a terminal (even with a solarized terminal).
(if window-system
    (progn
      (setq x-underline-at-descent-line t)

      ;; Don't change the font for some headings and titles
      (setq solarized-use-variable-pitch nil)

      (load-theme 'solarized-light t))
  (progn
    ;; Use an alternative solarized theme that handle terminals that are in
    ;; solarized color mode better. This is the theme from
    ;; ~/.emacs.d/local/themes/emacs-color-theme-solarized.
    ;;
    ;; The theme works fine for me in PuTTY but have some problems in urxvt.
    (load-theme 'solarized t)
    ;;(setq frame-background-mode 'light)
    ))

;; ----------------------------------------------------------------------------
;; external package: markdown-mode
;;   https://github.com/jrblevin/markdown-mode (Markdown documentation mode)
;; ----------------------------------------------------------------------------

(autoload 'markdown-mode "markdown-mode" "Major mode for editing Markdown files" t)
(add-to-list 'auto-mode-alist '("\\.markdown\\'" . markdown-mode))
(add-to-list 'auto-mode-alist '("\\.md\\'" . markdown-mode))
;;(add-hook 'markdown-mode-hook 'turn-on-auto-fill)

;; ----------------------------------------------------------------------------
;; external package: org
;;   http://orgmode.org (Org documentation mode)
;; ----------------------------------------------------------------------------

; Hide markup characters for bold, italics, underline etc.
(setq org-hide-emphasis-markers t)

; Word wrap mode.
(add-hook 'org-mode-hook 'visual-line-mode)

; Indent according to headline level.
(add-hook 'org-mode-hook 'org-indent-mode)

; Put all task logs into LOGBOOK drawer.
(setq org-log-into-drawer t)

; Log time stamp when marking a task as done.
(setq org-log-done 'time)

; Ask for log note when marking a task as done.
(setq org-log-done 'note)

; Prevent parent tasks from being marked as DUNE unless all child tasks are DONE.
(setq org-enforce-todo-dependencies t)

; Syntax highlight src blocks (I haven't gotten this to work yet).
(setq org-src-fontify-natively t)

; Don't emit validation link when exporting HTML
(setq org-html-validation-link nil)

;; org-agenda
(setq org-agenda-span 14)
(setq org-agenda-skip-scheduled-if-done t)
(setq org-agenda-skip-deadline-if-done t)
(setq org-deadline-warning-days 2)
(setq org-agenda-show-future-repeats 'next)

(setq org-agenda-files
      (quote
       ("~/org/agenda.org")))

;; org-capture
(setq org-default-notes-file "~/org/notes.org")

;; ----------------------------------------------------------------------------
;; external package: magit
;;   https://magit.vc/ (GIT plugin)
;; ----------------------------------------------------------------------------

(require 'magit)

;; ----------------------------------------------------------------------------
;; external package: auto-complete
;;   https://github.com/auto-complete/auto-complete
;; ----------------------------------------------------------------------------

(use-package auto-complete-config
             :ensure nil
	     :config
	     (setq-default ac-sources '())
	     (add-hook 'emacs-lisp-mode-hook 'ac-emacs-lisp-mode-setup)
	     (add-hook 'auto-complete-mode-hook 'ac-common-setup)
	     (global-auto-complete-mode t)
	     (setq
	      ac-menu-height 15
	      ac-ignore-case nil
	      ac-auto-start nil)
	     ;; Explicit key to trigger completion.
	     (define-key ac-mode-map (kbd "C-c k") 'auto-complete))

;; ----------------------------------------------------------------------------
;; external package: company
;;   http://company-mode.github.io/
;;
;;   Alternative auto complete package.
;; ----------------------------------------------------------------------------

(use-package company
	     :ensure nil
	     :config
	     (setq ;; Complete as you type behavior.
	      company-idle-delay nil
	      company-minimum-prefix-length 3
	      company-show-numbers 'left)
	     ;; Explicit key to trigger completion.
	     (define-key company-mode-map (kbd "C-c k") 'company-manual-begin))

;; ----------------------------------------------------------------------------
;; external package: go-mode
;;   https://github.com/dominikh/go-mode.el
;;
;;   Code navigation and completion is provided by Emacs lsp-mode together with
;;   the gopls LSP client. LSP mode is configured further down.
;;
;; install support tools:
;;   go get golang.org/x/tools/gopls@latest
;;   go get golang.org/x/tools/cmd/goimports@latest
;;   go get golang.org/x/tools/cmd/gorename@latest
;; ----------------------------------------------------------------------------

;; See configuration of LSP mode further down.
(require 'go-mode-autoloads)

(defun my/go-mode-hook ()
  ;; Disable auto-complete mode, use company mode.
  (auto-complete-mode -1))

(add-hook 'go-mode-hook 'my/go-mode-hook)

;; ----------------------------------------------------------------------------
;; Lisp
;; ----------------------------------------------------------------------------

;; TODO: This has not been tested for a while and I have not done much Lisp over
;;       the years. It's left in place because I have an ambition to pick up
;;       Lisp some day.

;; (setq inferior-lisp-program "/usr/bin/sbcl --noinform")

;; (autoload 'enable-paredit-mode "paredit" "Turn on pseudo-structural editing of Lisp code." t)
;; (add-hook 'emacs-lisp-mode-hook       #'enable-paredit-mode)
;; (add-hook 'eval-expression-minibuffer-setup-hook #'enable-paredit-mode)
;; (add-hook 'ielm-mode-hook             #'enable-paredit-mode)
;; (add-hook 'lisp-mode-hook             #'enable-paredit-mode)
;; (add-hook 'lisp-interaction-mode-hook #'enable-paredit-mode)
;; (add-hook 'scheme-mode-hook           #'enable-paredit-mode)

;; ;; Slime
;; (add-hook 'slime-repl-mode-hook (lambda () (paredit-mode +1)))

;; ;; Stop SLIME's REPL from grabbing DEL,
;; ;; which is annoying when backspacing over a '('
;; (defun override-slime-repl-bindings-with-paredit ()
;;   (define-key slime-repl-mode-map
;;     (read-kbd-macro paredit-backward-delete-key) nil))
;; (add-hook 'slime-repl-mode-hook 'override-slime-repl-bindings-with-paredit)

;; ----------------------------------------------------------------------------
;; OCaml
;;   https://ocaml.org/releases/
;;
;;   This config relies on OCaml opam packages being installed as merlin is
;;   quite dependent on the installed OCaml version.
;;
;; install support tools:
;;   opam update
;;   opam switch create 4.10.0
;;   opam install merlin tuareg utop ocp-indent odig
;;
;; ----------------------------------------------------------------------------

(let ((opam-share-dir (ignore-errors (car (process-lines "opam" "config" "var" "share")))))
  (when (and opam-share-dir (file-directory-p opam-share-dir))
    (let* ((opam-site-list-dir (concat opam-share-dir "/emacs/site-lisp"))
	   (opam-tuareg-site-file (format "%s/%s" opam-site-list-dir "tuareg-site-file")))

      ;; tuareg
      (when (file-exists-p (concat opam-tuareg-site-file ".el"))
	(load opam-tuareg-site-file))

      ;; ocp-indent
      (add-to-list 'load-path opam-site-list-dir)
      (require 'ocp-indent)

      ;; merlin
      (add-to-list 'load-path (expand-file-name "emacs/site-lisp" opam-share-dir))
      (autoload 'merlin-mode "merlin" nil t nil)
      (add-hook 'tuareg-mode-hook 'merlin-mode t)
      (add-hook 'caml-mode-hook 'merlin-mode t)
      (setq merlin-command 'opam))))

;; ----------------------------------------------------------------------------
;; external package: zig-mode
;;   https://github.com/ziglang/zig-mode (Zig language mode)
;; ----------------------------------------------------------------------------

(unless (version< emacs-version "24")
  (autoload 'zig-mode "zig-mode" nil t)
  (add-to-list 'auto-mode-alist '("\\.zig\\'" . zig-mode)))

;; ----------------------------------------------------------------------------
;; external package: nim-mode
;;   https://github.com/nim-lang/nim-mode (Nim language mode)
;; ----------------------------------------------------------------------------

(defun my/init-nim-mode ()
  "Local init function for `nim-mode'."

  ;; Make files in the nimble folder read only by default.
  ;; This can prevent to edit them by accident.
  (when (string-match "/\.nimble/" buffer-file-name) (read-only-mode 1))

  ;; If you want to experiment, you can enable the following modes by
  ;; uncommenting their line.
  (nimsuggest-mode 1)
  ;; Remember: Only enable either `flycheck-mode' or `flymake-mode' at the same time.
  ;; (flycheck-mode 1)
  ;; (flymake-mode 1)

  ;; The following modes are disabled for Nim files just for the case
  ;; that they are enabled globally.
  ;; Anything that is based on smie can cause problems.
  (auto-fill-mode 0)
  (electric-indent-local-mode 0))

(add-hook 'nim-mode-hook 'my/init-nim-mode)

;; ----------------------------------------------------------------------------
;; external package: lsp-mode
;;   https://github.com/emacs-lsp/lsp-mode (Language Server Protocol)
;;
;; useful key shortcuts:
;;
;;   see https://github.com/emacs-lsp/lsp-mode#commands for the full list.
;;
;; M-,          xref: Go back
;; M-.          xref: Find thing
;; M-?          xref: Find references
;; C-c l g g	Find definitions of the symbol under point.
;; C-c l g r	Find references of the symbol under point.
;; C-c l g i	Find implementations of the symbol under point.
;; C-c l g t	Find type definitions of the symbol under point.
;; C-c l g d	Find declarations of the symbol under point.
;; C-c l g h	Show the incoming call hierarchy for the symbol at point. (requires lsp-treemacs)
;; C-c l h h	Display the type signature and documentation of the thing at
;; C-c l r r	Rename the symbol (and all references to it).
;;
;; enabled for modes:
;;
;;   Go, see: https://github.com/golang/tools/blob/master/gopls/doc/emacs.md
;;
;; ----------------------------------------------------------------------------

(use-package lsp-mode
             :ensure nil
             :commands (lsp lsp-deferred)
             :hook ((go-mode . lsp-deferred)
                    (lsp-mode . (lambda ()
                                  (let ((lsp-keymap-prefix "C-c l"))
                                    (lsp-enable-which-key-integration)))))
             :config
             (progn
	       ;; May want to disable file watchers for big projects
               ;(setq lsp-enable-file-watchers nil)
               (define-key lsp-mode-map (kbd "C-c l") lsp-command-map))
             )

;; Set up before-save hooks to format buffer and add/delete imports.
;; Make sure you don't have other gofmt/goimports hooks enabled.
(defun lsp-go-install-save-hooks ()
  (add-hook 'before-save-hook #'lsp-format-buffer t t)
  (add-hook 'before-save-hook #'lsp-organize-imports t t))
(add-hook 'go-mode-hook #'lsp-go-install-save-hooks)

;; Provides fancier overlays.
(use-package lsp-ui
	     :ensure nil
	     :commands lsp-ui-mode
	     :bind (:map lsp-ui-mode-map
;;			 ("M-."     . lsp-ui-peek-find-definitions)
;;			 ("M-?"     . lsp-ui-peek-find-references)
			 ("C-c d" . lsp-ui-doc-glance)
			 ("C-c m" . lsp-ui-imenu)
			 )
	     :config
	     (setq lsp-ui-doc-enable nil
		   lsp-ui-sideline-enable nil
		   lsp-ui-peek-peek-height 35
		   ;; word selection does not work in the left pane without 'always
		   lsp-ui-peek-fontify 'always
		   lsp-ui-flycheck-enable t)
	     :after lsp-mode)

;; M-x lsp-treemacs-symbols is probably the most useful option that this enables.
;; M-x treemacs is also gold to get a project like directory browser.
(use-package lsp-treemacs)

;; company-lsp integrates company mode completion with lsp-mode.
;; completion-at-point also works out of the box but doesn't support snippets.
(use-package company-lsp
	     :ensure nil
	     :commands company-lsp)

;; Provides snippet support, also used to fill in arguments after function name
;; auto completion.
(use-package yasnippet
	     :ensure nil
	     :commands yas-minor-mode
	     :hook
	     ((go-mode c-mode-common) . yas-minor-mode))

;; ----------------------------------------------------------------------------
;; external package: ccls
;;   https://github.com/MaskRay/emacs-ccls/ (C/C++ LSP indexer extensions)
;;
;; install support tools:
;;   https://github.com/MaskRay/ccls
;; ----------------------------------------------------------------------------

(let ((exe-path (bound-and-true-p my/ccls-exe-path)))
  (when (and exe-path (file-exists-p exe-path))
    (use-package ccls
      :ensure nil
      :hook ((c-mode c++-mode objc-mode cuda-mode) .
	     (lambda () (require 'ccls) (lsp)))
      :config
      (setq ccls-executable exe-path))))

;; ----------------------------------------------------------------------------
;; Automatic window size adjustment
;; ----------------------------------------------------------------------------

(defun set-frame-size-according-to-resolution ()
  (interactive)
  (if window-system
  (progn
    ;; use 120 char wide window for largeish displays
    ;; and smaller 80 column windows for smaller displays
    ;; pick whatever numbers make sense for you
    (if (> (x-display-pixel-width) 1280)
	   (add-to-list 'default-frame-alist (cons 'width 130))
	   (add-to-list 'default-frame-alist (cons 'width 80)))
    ;; for the height, subtract some pixelsfrom the screen height
    ;; (for panels, menubars and whatnot), then divide by the height
    ;; of a char to get the height we want
    (add-to-list 'default-frame-alist
	 (cons 'height (/ (- (x-display-pixel-height) 100)
			     (frame-char-height)))))))

(set-frame-size-according-to-resolution)
